#pragma once

#include "fault.h"
#include "lut.h"

using ll = long long;

namespace atpg_ls {

struct TMP_FAULT {
    /* data */
    Gate *g;

    std::string name;
    int stuck_at;
    int is_stem;
    int is_PO;

    friend bool operator < (const TMP_FAULT &lhs, const TMP_FAULT &rhs) {
        if(lhs.name != rhs.name) {
            return lhs.name < rhs.name;
        }
        if(lhs.stuck_at != rhs.stuck_at) {
            return lhs.stuck_at < rhs.stuck_at;
        }
        if(lhs.is_stem != rhs.is_stem) {
            return lhs.is_stem < rhs.is_stem;
        }
        if(lhs.is_PO != rhs.is_PO) {
            return lhs.is_PO < rhs.is_PO;
        }
        return false;
    }
};

class Simulator;
class Circuit;

class LUTCircuit {
public:
std::vector<LUT*> PIs;
std::vector<LUT*> POs;
std::vector<LUT*> luts;
std::vector<LUT*> rtopo_luts;

int** fault_detected;

void print();

// local search
void ls_update(std::vector<LUT*> &unsat);
LUT* ls_pick();
void ls_flip(LUT *lut);
void ls_main();
void ls_init();
void ls_random_sol();
void ls_gen_sol(Gate* target, int stuck_at);

// checker
double check();

Simulator *simulator;
Circuit *C;

int step;

};

class Circuit {
public:
std::vector<Gate*> PIs;
std::vector<Gate*> POs;
std::vector<Gate*> gates;
std::vector<Gate*> rtopo_gates;

std::unordered_map<std::string, Gate*> name2gate;

void parse_from_file(const char *filename);
void insert_lines_for_stem();

void print();

LUTCircuit* build_lut_circuit();

void init_topo_index();
void init_avg_dist();

};

}