#include <set>
#include <chrono>
#include <iostream>

#include "circuit.h"
#include "paras.h"
#include "simulator.h"
#include "sat_atpg.h"

using namespace atpg_ls;

LUT* LUTCircuit::ls_pick() {

    LUT* pick = nullptr;
    
    // 采样
    for(int i=0; i<OPT(t); i++) {
        LUT* lut = luts[rand() % luts.size()];
        if(!lut->CC) continue;
        lut->cal_score();
        if(lut->score <= 0) continue;
        if(pick == nullptr || lut->score > pick->score) {
            pick = lut;
        }
    }

    if(pick != nullptr) {
        return pick;
    }

    std::vector<LUT*> unsat;

    // 动态加权
    ls_update(unsat);

    if(unsat.size() == 0) return nullptr;

    return unsat[rand() % unsat.size()];
}

void LUTCircuit::ls_flip(LUT *lut) {

    lut->CC = 0;
    lut->uptag = 0;
    lut->flip_value();
    if(!lut->isPI) lut->vsat = !lut->vsat;

    for(LUT* out : lut->fanouts) {
        out->CC = 1;
        out->vsat = (out->cal_value() == out->value);
    }

    lut->cal_fault_info(lut->fd, lut->fpl);

    for(LUT* r : lut->reigon) {
        r->CC = 1;
        int t_fd[2], t_fpl[2];
        r->cal_fault_info(t_fd, t_fpl);

        if(t_fd[0] != r->fd[0] || t_fd[1] != r->fd[1] || t_fpl[0] != r->fpl[0] || t_fpl[1] != r->fpl[1]) {
            r->uptag = 1;
        }
    }
}

void LUTCircuit::ls_update(std::vector<LUT*> &unsat) {

    // printf("?????????????????ls_update\n");

    // if(rand() % 10000 <= OPT(sp)) {

    // } else {

        for(LUT* lut : luts) {
            if(!lut->vsat) {

                // printf("ocost: %d add: %d\n", lut->vunat_cost);
                if(lut->vunat_cost + OPT(vsat_inc) > OPT(vsat_max)) {
                    lut->vunat_cost = OPT(vsat_max);
                } else {
                    lut->vunat_cost += OPT(vsat_inc);
                }
                // printf("cost: %d add: %d\n", lut->vunat_cost, OPT(vsat_inc));
                
                // for(LUT* out : lut->fanouts) {
                //     if(out->vunat_cost -= OPT(vsat_inc) > 1) {
                //         out->vunat_cost -= OPT(vsat_inc);
                //     } else {
                //         out->vunat_cost = 1;
                //     }
                // }
                
                unsat.push_back(lut);
            }
            if(lut->uptag) {
                if(lut->up_cost + OPT(up_inc) > OPT(up_max)) {
                    lut->up_cost = OPT(up_max);
                } else {
                    lut->up_cost += OPT(up_inc);
                }
            }
            for(int i=lut->fanins.size(); i<lut->fanins.size()+lut->inner_gates.size(); i++) {
                Gate* g = lut->inner_gates[i-lut->fanins.size()];
                LUT::FaultInfo &fi = lut->fault_table[i][lut->input_var];
                if(!fi.fd[0] && g->fault_detected_weight[0] > 0 && g->fault_detected_weight[0] + OPT(fw_inc) <= OPT(fw_max)) {
                    g->fault_detected_weight[0] += OPT(fw_inc);
                }
                if(!fi.fd[1] && g->fault_detected_weight[1] > 0 && g->fault_detected_weight[1] + OPT(fw_inc) <= OPT(fw_max)) {
                    g->fault_detected_weight[1] += OPT(fw_inc);
                }
                if(!fi.fd[0] && g->fault_propagated_weight[0] > 0 && g->fault_propagated_weight[0] + OPT(fw_inc) <= OPT(fw_max)) {
                    g->fault_propagated_weight[0] += OPT(fw_inc);
                }
                if(!fi.fd[1] && g->fault_propagated_weight[0] > 0 && g->fault_propagated_weight[1] + OPT(fw_inc) <= OPT(fw_max)) {
                    g->fault_propagated_weight[1] += OPT(fw_inc);
                }
            }
        }
    // }
}

void LUTCircuit::ls_main() {

    printf("====== local search start ======\n");

    printf("initing lookup table ...\n");
    ls_init();

    printf("ganerating fault info ...\n");

    int num_gates = 0;
    int num_total_fault = 0;
    int num_detected_fault = 0;
    int num_undetected_fault = 0;
    int num_pattern = 0;

    for(LUT* lut : luts) {
        num_gates += lut->inner_gates.size();
    }
    assert(num_gates == C->gates.size());
    fault_detected = new int*[num_gates];
    for(int i=0; i<num_gates; i++) {
        fault_detected[i] = new int[2];
        fault_detected[i][0] = fault_detected[i][1] = 0;
    }
    num_total_fault = num_undetected_fault = num_gates * 2;

    printf("staring local search ...\n");
    
    std::queue<TMP_FAULT> faults;

    for(Gate* g : C->gates) {

        std::string name = g->name;
        // if(name.find("_") != std::string::npos) {
        //     name = name.substr(0, name.find("_"));
        // }

        if(g->is_stem) {
            // printf("stem: %s\n", name.c_str());
            faults.push(TMP_FAULT{g, name, 0, 1, 0});
            faults.push(TMP_FAULT{g, name, 1, 1, 0});
        } else {

            if(g->isPO) {
                // printf("out: %s\n", name.c_str());
                faults.push(TMP_FAULT{g, name, 0, g->type != Gate::LINE, g->type == Gate::LINE});
                faults.push(TMP_FAULT{g, name, 1, g->type != Gate::LINE, g->type == Gate::LINE});
                continue;
            }

            assert(g->fanouts.size() == 1);
            Gate *fanout = g->fanouts[0];

            int stem = (!g->isPI) && (g->type != Gate::LINE);

            if(fanout->type == Gate::Type::BUF || fanout->type == Gate::Type::NOT) {
                continue;
            } else if(fanout->type == Gate::Type::XOR || fanout->type == Gate::Type::XNOR) {
                faults.push(TMP_FAULT{g, name, 0, stem, 0});
                faults.push(TMP_FAULT{g, name, 1, stem, 0});
            } else if(fanout->type == Gate::Type::NAND || fanout->type == Gate::Type::AND) {
                faults.push(TMP_FAULT{g, name, 1, stem, 0});
            } else if(fanout->type == Gate::Type::NOR || fanout->type == Gate::Type::OR) {
                faults.push(TMP_FAULT{g, name, 0, stem, 0});
            } else {
                assert(false);
            }
        }
    }

    printf("fault-size: %d\n", faults.size());

    std::vector<TMP_FAULT> t_faults;
    std::set<TMP_FAULT> t_verify_set;

    while(!faults.empty()) {
        TMP_FAULT f = faults.front();
        faults.pop();

        assert(t_verify_set.count(f) == 0);

        t_faults.push_back(f);
        t_verify_set.insert(f);
    }

    for(TMP_FAULT &f : t_faults) {
        faults.push(f);
    }

    // std::sort(t_faults.begin(), t_faults.end(), [](const TMP_FAULT &a, const TMP_FAULT &b) {
    //     int ta = std::atoi(a.g->name.substr(1).c_str());
    //     int tb = std::atoi(b.g->name.substr(1).c_str());
    //     return ta < tb;
    // });

    // for(auto &f : t_faults) {
        
    // }

    printf("fault-size: %d\n", t_faults.size());
    printf("verify-size: %d\n", t_verify_set.size());
    assert(t_faults.size() == t_verify_set.size());

    // for(TMP_FAULT &f : t_faults) {

    //     printf("!! Fault: %s stuck-at: %d is_stem: %d is_po: %d\n", f.g->name.c_str(), f.stuck_at, f.is_stem, f.is_PO);
        
    //     std::vector<int> input_vector;
    //     bool detected = sat_atpg(f, input_vector);

    //     printf(">> Fault: %s stuck-at: %d is_stem: %d is_po: %d detected: %d\n", f.g->name.c_str(), f.stuck_at, f.is_stem, f.is_PO, detected);
        
    // }

    // exit(0);

    auto start = std::chrono::high_resolution_clock::now();

    // int cnt = 0;

    while(!faults.empty()) {

        while(!faults.empty()) {
            TMP_FAULT f = faults.front();
            if(fault_detected[f.g->id-1][f.stuck_at]) {
                faults.pop();
            } else {
                break;
            }
        }

        if(faults.empty()) break;

        ls_random_sol();

        TMP_FAULT f = faults.front(); faults.pop();
        Gate* g = f.g;
        int stuck_at = f.stuck_at;

        printf("start with fault: %s SA%d\t", g->name.c_str(), stuck_at);

        std::vector<int> inputs;

        printf("verify ...");

        int res1 = sat_atpg(f, inputs);
        
        if(res1 == 0) {
            printf(" unsat!\n");
            continue;
        }

        assert(inputs.size() == PIs.size());

        for(int i=0; i<inputs.size(); i++) {
            PIs[i]->value = inputs[i];
        }

        printf("sim: %s %d\n", g->name.c_str(), stuck_at);
        for(int i=0; i<inputs.size(); i++) {
            printf("%d ", inputs[i]);
        }
        printf("\n");

        int score;
        simulator->simulate(PIs, score, fault_detected);

        if(simulator->name2gate[g->name]->fault_detected[stuck_at]) {
            printf(" successful!\n");

            for(LUT* lut : luts) {
                lut->input_var = 0;
                for(int i=0; i<lut->fanins.size(); i++) {
                    LUT* in = lut->fanins[i];
                    lut->input_var |= (in->value << i);
                }
            }

            ls_gen_sol(g, stuck_at);

            int res = simulator->verify(this, fault_detected);
            assert(res == 1);

        } else {
            printf(" failed!\n");
            simulator->name2gate[g->name]->fault_detected[stuck_at] = 1;
            // exit(-1);
        }

        // printf("g: %s\n", g->name.c_str());
        // assert(simulator->name2gate[g->name]->fault_detected[stuck_at]);

        int num_fault = 0;

        for(Gate* g : simulator->gates) {
            if(g->fault_detected[0] && !fault_detected[g->id-1][0]) {
                fault_detected[g->id-1][0] = 1;
                num_detected_fault++;
                num_undetected_fault--;
                num_fault++;
            }

            if(g->fault_detected[1] && !fault_detected[g->id-1][1]) {
                fault_detected[g->id-1][1] = 1;
                num_detected_fault++;
                num_undetected_fault--;
                num_fault++;
            }
        }

        if(num_fault > 0) {
            num_pattern++;
        }

        auto end = std::chrono::high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count();

        // cnt++;
        printf("Cover: %.2f%% pattern: %d new_detected: %d undected: %d time: %.2fs\n", (double)num_detected_fault / num_total_fault * 100, num_pattern, num_fault, num_undetected_fault, (double)duration/1000);

        // break;
    }

    auto end = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count();
    std::cout << "Execution time: " << duration << " milliseconds" << std::endl;
}

void LUTCircuit::ls_gen_sol(Gate* target, int stuck_at) {
    
    for(int step=0; ; step++) {
        // printf("step: %d\n", step);

        LUT* pick = ls_pick();

        if(pick == nullptr) {
            break;
        }

        // pick->cal_score();

        // double t1 = check();

        // printf(">>>>>>>>>>>>>\n");
        // printf("dert_score: %.2f\n", pick->score);
        // printf("dert_unsat_cost: %.2f\n", pick->score_value_unsatisfied_cost);
        // printf("dert_fault_detected_weight: %.2f\n", pick->score_fault_detected_weight);
        // printf("dert_fault_propagated_weight: %.2f\n", pick->score_fault_propagated_weight);
        // printf("dert_up_cost: %.2f\n", pick->score_fault_update_cost);

        int last = simulator->name2gate[target->name]->fault_detected[stuck_at];

        ls_flip(pick);

        // double t2 = check();
        // assert(((t2 - t1) - pick->score) < 1e-6);

        if(pick->isPI) {
            int score;
            simulator->simulate(PIs, score, fault_detected);
            if(simulator->name2gate[target->name]->fault_detected[stuck_at] != last) {
                ls_flip(pick);
                pick->CC = 0;
            }
            // printf("step: %d fd: %d\n", step, score);
        }
    }
}

void LUTCircuit::ls_init() {
    for(LUT* lut : luts) {
        static int cnt = 0;
        printf("[%d/%d]\n", ++cnt, luts.size());
        lut->init_lookup_table();
    }

    for(LUT* lut : luts) {
        for(LUT* out :lut->fanouts) {
            for(int i=0; i<out->fanins.size(); i++) {
                if(out->fanins[i] == lut) {
                    lut->fault_info.push_back(out->fault_table[i]);
                }
            }
        }
    }

    for(LUT* lut : luts) {
        std::set<LUT*> t_reigon;
        for(LUT* out : lut->fanouts) {
            for(LUT* in : out->fanins) {
                t_reigon.insert(in);
            }
        }
        for(LUT* in : lut->fanins) {
            t_reigon.insert(in);
        }
        t_reigon.erase(lut);
        for(LUT* r : t_reigon) {
            lut->reigon.push_back(r);
        }
    }

    for(LUT* lut : luts) {
        std::set<LUT*> t_reigon;
        for(LUT* r: lut->reigon) {
            t_reigon.insert(r);
            for(LUT* out2 : r->fanouts) {
                t_reigon.insert(out2);
                for(LUT* in2 : out2->fanins) {
                    t_reigon.insert(in2);
                }
            }
        }
        for(LUT* r : lut->fanouts) {
            t_reigon.insert(r);
        }
        for(LUT* r : t_reigon) {
            lut->update_reigon.push_back(r);
        }
    }
}

void LUTCircuit::ls_random_sol() {

    step = 0;

    std::vector<int*> t_focus;

    for(LUT* lut : luts) {
        lut->up_cost = 1;

        lut->vunat_cost = 1;
        if(lut->isPI) {
            lut->vunat_cost = OPT(vsat_inc) * 100;
        }

        lut->CC = 1;
        for(Gate* g : lut->inner_gates) {

            g->fault_detected_weight[0] = !fault_detected[g->id-1][0];
            g->fault_detected_weight[1] = !fault_detected[g->id-1][1];

            if(g->fault_detected_weight[0]) {
                t_focus.push_back(&g->fault_detected_weight[0]);
            }

            if(g->fault_detected_weight[1]) {
                t_focus.push_back(&g->fault_detected_weight[1]);
            }
            
            g->fault_propagated_weight[0] = !fault_detected[g->id-1][0];
            g->fault_propagated_weight[1] = !fault_detected[g->id-1][1];
        }
    }

    int *tw = t_focus[rand()%t_focus.size()];
    *tw = 100000;
    
    for(LUT* lut : luts) {
        lut->uptag = 0;
        lut->value = rand() % 2;
    }

    for(LUT* lut : luts) {
        lut->input_var = 0;
        for(int i=0; i<lut->fanins.size(); i++) {
            LUT* in = lut->fanins[i];
            lut->input_var |= (in->value << i);
        }
    }

    for(LUT* lut : luts) {
        lut->vsat = (lut->value == lut->cal_value());
    }

    for(LUT* lut : rtopo_luts) {
        lut->cal_fault_info(lut->fd, lut->fpl);
    }

    for(LUT* lut : luts) {
        lut->cal_score();
    }
}